(* Time-stamp: <modified the 12/04/2013 (at 08:46) by Erwan Jahier> *)

open Lxm
open Lv6errors
open AstV6
open AstCore

(* get the first package in the package/model list *)

let (doit : Lv6MainArgs.t -> AstV6.pack_or_model list -> Ident.idref option -> LicPrg.t) = 
  fun opt srclist main_node ->
    let syntax_tab = AstTab.create srclist in
    (* Pour chaque package, on a un solveur de rfrences
       globales, pour les types, const et node :
       - les rfrences pointes (p::n) sont recherches
       directement dans la syntax_tab puisqu'il n'y a pas 
       d'ambiguit
       - les rfrences simples sont recherches :
       . dans le pack lui-mme
       . dans un des packs dclars "uses", avec
       priorit dans l'ordre
    *)
    let lic_tab = LicTab.create syntax_tab in
    Verbose.exe ~level:2 (fun () -> AstTab.dump syntax_tab);

    let lic_tab = match main_node with
      | None -> LicTab.compile_all lic_tab
      | Some main_node -> 
        if opt.Lv6MainArgs.compile_all_items then
          LicTab.compile_all lic_tab
        else 
          LicTab.compile_node lic_tab main_node
    in
    let zelic = LicTab.to_lic_prg lic_tab in
    (* limination polymorphisme  surcharge *)
    let zelic = L2lRmPoly.doit zelic in
    (* alias des types array *)
    (*     let zelic = L2lAliasType.doit zelic in *)
    let zelic = if not opt.Lv6MainArgs.inline_iterator then zelic else
        (* to be done before array expansion otherwise they won't be expanded *)
        L2lExpandMetaOp.doit zelic 
    in
    let zelic = 
      if 
        Lv6MainArgs.global_opt.Lv6MainArgs.one_op_per_equation 
        || opt.Lv6MainArgs.expand_nodes (* expand performs no fixpoint, so it will work
                                   only if we have one op per equation...*)
      then 
        (* Split des equations (1 eq = 1 op) *)
        L2lSplit.doit zelic 
      else 
        zelic
    in
    (* Array and struct expansion: to do after polymorphism elimination *)
    let zelic = if not opt.Lv6MainArgs.expand_nodes then zelic else 
        L2lExpandNodes.doit opt.Lv6MainArgs.dont_expand_nodes zelic 
    in
    let zelic = if not opt.Lv6MainArgs.expand_arrays then zelic else
        L2lExpandArrays.doit zelic 
    in    

    (* Currently only works in this mode *)
    if Lv6MainArgs.global_opt.Lv6MainArgs.ec then L2lCheckLoops.doit zelic;
    L2lCheckOutputs.doit zelic;

    zelic
      
let test_lex ( lexbuf ) = (
  let tk = ref (Lv6lexer.lexer lexbuf) in 
    while !tk <> Lv6parser.TK_EOF do
      match (Lv6lexer.token_code !tk) with 
	  ( co , lxm ) ->
	    Printf.printf "line %3d col %2d to %2d : %15s = \"%s\"\n"
	      (line lxm) (cstart lxm) (cend lxm) co (str lxm) ;
	    tk := (Lv6lexer.lexer lexbuf)
    done
)

(* Retourne un AstV6.t *)
let lus_load lexbuf = 
  let tree = Lv6parser.program Lv6lexer.lexer lexbuf in
    LicName.update_fresh_var_prefix ();
    (* ICI  *)
    AstRecognizePredef.f tree
  
type maybe_packed = 
  | Packed of AstV6.pack_or_model
  | Unpacked of AstV6.packbody 

let (get_source_list : Lv6MainArgs.t -> string list -> AstV6.pack_or_model list) =
  fun opt infile_list -> 
    let (get_one_source : string -> string list * maybe_packed list) = 
      fun infile -> 
        let incl_files, l =
          let lexbuf = Lv6MainArgs.lexbuf_of_file_name infile in
          if opt.Lv6MainArgs.tlex then test_lex lexbuf;
          match (lus_load lexbuf) with
            | PRPackBody(incl_files, pbdy) -> incl_files, [Unpacked pbdy]
            | PRPack_or_models(incl_files, nsl) -> incl_files, (List.map (fun ns -> Packed ns) nsl)
        in
        (* If included files have a relative path, strange things may happen.
           Hence we make the path absolute, using the directory of the includer.
        *)
        let includer_dir = Filename.dirname infile in
        let fix_dir f = if Filename.is_relative f then Filename.concat includer_dir f else f in
        let incl_files = List.map fix_dir incl_files in
        incl_files, l
    in
    let rec (get_remaining_source_list : maybe_packed list * string list * string list -> 
             maybe_packed list * string list * string list) =
      fun (pack_acc, compiled, to_be_compiled) -> 
        match to_be_compiled with
          | [] -> (pack_acc, compiled, [])
          | infile::tail ->
            let infile = FilenameExtras.simplify infile in
            if List.mem infile compiled then
              get_remaining_source_list (pack_acc, compiled, tail)
            else
              let included_files, pack = get_one_source infile in
              let new_pack_acc = pack_acc@pack in
              get_remaining_source_list(
                new_pack_acc, 
                infile::compiled, 
                tail@included_files)
    in
    let infile_list = 
      (* We need absolute paths to make sure that files are not
         included several times.  Indeed, otherwise,
         FilenameExtras.simplify may miss some simplifications.  For
         example, consider the files "../../x/toto.lus" and
         "../toto.lus".  They actually refer to the same file if the
         current directory is a sub-directory of "x". Working with
         absolute paths solves the problem.
         
      *)
      let make_it_absolute f = 
        if Filename.is_relative f then Filename.concat (Sys.getcwd ()) f else f 
      in
      List.map make_it_absolute infile_list
    in
    let first_file = assert (infile_list <> []); List.hd infile_list in
    let included_files, first_pack = get_one_source first_file in
    let (pack_list, _compiled_files, included_files) = 
      get_remaining_source_list (first_pack, [first_file], (List.tl infile_list) @ included_files)
    in
    let _ = assert (included_files=[]) in
    let packed_list, unpacked_list = 
      List.fold_left 
        (fun (pl, upl) p -> 
          match p with
            | Packed p ->  p::pl, upl
            | Unpacked up -> pl, up::upl
        )
        ([], [])
        pack_list
    in
    let unpacked_merged_opt = (* All unpacked files are merged into one single package *)
      List.fold_left
        (fun acc pbody -> 
          match acc with
            | None -> Some pbody
            | Some pbody_acc -> 
              let add tbl x y =
                     (* Let's perform some clashes checks *)
                if Hashtbl.mem tbl x then
                  let ybis = Hashtbl.find tbl x in
		            print_string ("*** Error: "^(Ident.to_string x)^" is defined twice: \n\t" ^ 
                                   (Lxm.details y.src) ^ "\n\t" ^
                                   (Lxm.details ybis.src) ^ ".\n"); 
                  exit 2
                else
                  Hashtbl.add tbl x y
              in
              Hashtbl.iter (fun x y -> add pbody_acc.pk_const_table x y) pbody.pk_const_table;
              Hashtbl.iter (fun x y -> add pbody_acc.pk_type_table x y) pbody.pk_type_table;
              Hashtbl.iter (fun x y -> add pbody_acc.pk_node_table x y) pbody.pk_node_table;
              Some { pbody_acc with
                pk_def_list=pbody_acc.pk_def_list@pbody.pk_def_list;
              }
        )
        None
        unpacked_list
    in
    match unpacked_merged_opt with
      | None -> packed_list
      | Some unpacked_merged ->
        let name = 
          try Filename.chop_extension (Filename.basename first_file) 
          with _ -> 
		      print_string ("*** '"^first_file^"': bad file name.\n"); exit 1
        in
        let pi = AstV6.give_pack_this_name (Ident.pack_name_of_string name) unpacked_merged in
        let p = NSPack (Lxm.flagit pi (Lxm.dummy name)) in
        p::packed_list
