(* Time-stamp: <modified the 12/04/2013 (at 08:47) by Erwan Jahier> *)

open Verbose
open AstV6
open AstCore
open Lxm
open Lv6errors
open Parsing
open Format

open Lv6MainArgs
let my_exit opt i =
  close_out opt.oc;
  if Sys.file_exists opt.outfile then Sys.remove opt.outfile;
  exit i

let rec first_pack_in =
  function
    | (AstV6.NSPack pi)::_ -> pi.it.pa_name
    | (AstV6.NSModel _)::tail -> first_pack_in tail
    | [] -> raise (Global_error "No package has been provided")

let main  = (
  (* Compile.init_appli () ; *)
  (* parse_args (); *)
  let opt = Lv6MainArgs.parse Sys.argv in
  Verbose.exe ~level:3 (fun () ->
    Gc.set { (Gc.get ()) with Gc.verbose = 0x01 }
  );
  if opt.run_unit_test then (
    UnifyType.unit_test ();
    exit 0
  );
  if (opt.infiles = []) then (
    Lv6MainArgs.usage stderr opt;
    exit 1
  );
  let new_dft_pack = Filename.basename (Filename.chop_extension (List.hd opt.infiles)) in
  Ident.set_dft_pack_name new_dft_pack;

  let main_node = 
    if opt.main_node = "" then None else 
      Some (Ident.idref_of_string opt.main_node)
  in
  if opt.outfile <> "" then opt.oc <- open_out opt.outfile;
  (try (
    let nsl = Compile.get_source_list opt opt.infiles in
    let lic_prg = Compile.doit opt nsl main_node in

    if opt.exec then
      (match main_node with
        | None -> (
          let first_file = List.hd opt.infiles in
          let name = 
            try Filename.chop_extension (Filename.basename first_file) 
            with _ -> 
		        print_string ("*** '"^first_file^"': bad file name.\n"); exit 1
          in
          let nk = (Lic.node_key_of_idref (Ident.to_idref name)) in
          if LicPrg.node_exists lic_prg nk then (
            print_string ("WARNING: No main node is specified. I'll try with " ^ name ^"\n");
            flush stdout;
            let msk, zesoc = Lic2soc.f lic_prg nk in
            SocExec.f opt zesoc msk
          ) else (
            print_string ("Error: no node is specified, cannot exec.\n");
            flush stdout;
            exit 1
          )
        )
        | Some main_node -> 
          let msk, zesoc = Lic2soc.f lic_prg (Lic.node_key_of_idref main_node) in
          SocExec.f opt zesoc msk
      ) else (   
        LicPrg.to_file opt lic_prg main_node
       );

    Verbose.exe ~level:3 (fun () -> Gc.print_stat stdout);
  ) with 
      Sys_error(s) ->
        prerr_string (s^"\n");
        my_exit opt 1
    | Global_error s ->
      print_global_error s ;
      my_exit opt 1
    | Parse_error ->
      print_compile_error (Lxm.last_made ()) "syntax error";
      my_exit opt 1
    | Unknown_var(lxm,id) -> 
      print_compile_error lxm ("unknown variable (" ^ (Ident.to_string id) ^")")
    | Unknown_constant(lxm,str) -> 
      print_compile_error lxm ("unknown constant (" ^ str ^")")
    | Compile_error(lxm,msg) -> 
      print_compile_error lxm msg;
      my_exit opt 1
    | L2lCheckLoops.Error(lxm,msg,lic_prg) -> 
      (* Sometime it helps to see the current state of the faulty program *)
      LicPrg.to_file opt lic_prg main_node;
      flush  opt.oc;
      print_compile_error lxm msg;
      my_exit opt 1
    | Assert_failure (file, line, col)  -> 
      prerr_string (
        "\n*** oops: lus2lic internal error\n\tFile \""^ file ^
          "\", line " ^ (string_of_int line) ^ ", column " ^
          (string_of_int col) ^ "\n*** when compiling lustre program" ^
          (if List.length opt.infiles > 1 then "s " else " ") ^
          (String.concat ", " opt.infiles) ^ "\n"^
          "\n*** You migth want to sent a bug report to "^Lv6version.maintainer ^"\n") ; 
      my_exit opt 2
  );
(* | Compile_node_error(nkey,lxm,msg) -> ( *)
(* print_compile_node_error nkey lxm msg ; *)
(* exit 1 *)
(* ) *)
(* | Global_node_error(nkey,msg) -> ( *)
(* print_global_node_error nkey msg ; *)
(* exit 1 *)
(* ) *)

    close_out opt.oc
        
);
