(*-----------------------------------------------------------------------
** Copyright (C) - Verimag.
** This file may only be copied under the terms of the GNU Library General
** Public License
**-----------------------------------------------------------------------
**
** File: rif.ml
** Author: jahier@imag.fr
*)


open List

let lexer = Genlex.make_lexer ["q"; "#"; "x"; "load_luc"; "#@"; "@#"]

(* xxx Which pragmas should be defined ? *)

let rif_pragmas = ["inputs"]
(* let rif_pragmas = ["outs";"outputs";"program";"inputs";"step";"reset"] *)


type stream = Genlex.token Stream.t

let rec (parse_string_list : stream -> string list -> string list) =
  fun stream sl ->
    try (
      match (Stream.next stream) with
          (Genlex.String(_, str)) -> parse_string_list stream (str::sl)
        | _  -> failwith ("### rif parse error. A \"string\" (wrapped with double" ^
                          "quotes) was expected. \n")
    )
    with Stream.Failure ->
      sl

open Data

(*------------------------------------------------------------------------*)

let debug = false
let read_line label ic oc =
  let str = input_line ic in
  let _ = if debug then (print_string (label^ "read_line:"^str^"\n"); flush stdout) in
  let str = Str.global_replace (Str.regexp "\013") "" str in
    (match oc with
      | Some oc -> output_string oc str; flush oc
      | None -> ());
    str

let get_stream ic oc = 
  try lexer (Stream.of_string (read_line "" ic oc))
  with e  ->
    print_string ("*** Error when parsing RIF: " ^ (Printexc.to_string e) ^ "\n");
    flush stdout;
    exit 2


let (rm_blank : string -> string) =
  fun s -> 
    let buff = ref "" in
      for i = 0 to String.length s - 1 do
        match s.[i] with
          | ' ' | '\t' | '\n' | '\"' -> ()
          | c -> buff:=!buff^(String.make 1 c)
      done;
      !buff        

let (to_pair : string -> string * string) =
  fun s ->
    match Str.split (Str.regexp ":") s with
      | [n;t] -> rm_blank n, rm_blank t
      | _ -> failwith ("Rif internal error: Cannot split '"^s^"'. It should contain exactly one ':'")


let _ = assert (to_pair "T:bool" = ("T","bool"))


let rec (read_until_pragma_end :  in_channel -> out_channel option -> string -> string) =
  fun ic oc str -> 
    let line = read_line "" ic oc in
      try
        let i = Str.search_forward (Str.regexp "#@") line 0 in
          (
            String.sub line 0 i) ^ str
      with Not_found -> 
        (*         let _ = print_string ("read_until_pragma_end done: '"^str^"'\n"); flush stdout   in *)
        
(*         let _ = print_string ("read_until_pragma_end adding: "^line^"\n"); flush stdout   in *)
          read_until_pragma_end ic oc (str^" "^line)

(* exported *)
let rec (read_interface : string -> in_channel -> out_channel option -> 
          (string * string) list * (string * string) list) =
  fun label ic oc -> 
(*     let _ = print_string ("read_interface\n"); flush stdout in *)
    let rec loop ins outs =
      if ins <> [] && outs <> [] then ins, outs else
      let line = read_line label ic oc in
        try
          if Str.string_match (Str.regexp "#inputs") line 0 then
            let str = String.sub line 7 (String.length line - 7) in
            let l = Str.split (Str.regexp " ") str in
            let l = List.filter (fun str -> str <> "") l in
              loop (List.map to_pair l) outs
                
          else if Str.string_match (Str.regexp "@#inputs") line 0 then          
            let str = String.sub line 8 (String.length line - 8) in
            let str = read_until_pragma_end ic oc str in
            let l = Str.split (Str.regexp " ") str in
            let l = List.filter (fun str -> str <> "") l in
              loop (List.map to_pair l) outs
                
          else if Str.string_match (Str.regexp "#outputs") line 0 then
            let str = String.sub line 8 (String.length line - 8) in
            let l = Str.split (Str.regexp " ") str in
            let l = List.filter (fun str -> str <> "") l in
              loop ins (List.map to_pair l)
                
          else if Str.string_match (Str.regexp "@#outputs") line 0 then          
            let str = String.sub line 9 (String.length line - 9) in
            let str = read_until_pragma_end ic oc str in
            let l = Str.split (Str.regexp " ") str in
            let l = List.filter (fun str -> str <> "") l in
              loop ins (List.map to_pair l)
          else
            loop ins outs
        with _ -> 
          print_string "*** Error when parsing socket content: I was expecting #inputs/#outputs\n";
          print_string ("*** And I've read '" ^ line ^"'\n");
          flush stdout;
          loop ins outs
    in
      loop [] []

exception Bye

(* exported *)
let rec (read :  ?pragma:(string list) -> in_channel -> out_channel option -> vntl -> subst list) =
  fun ?(pragma = []) ic oc vntl  ->
    (**  Reads input values on ic. It should follow the rif format. *)
    let tbl = [] in
    if vntl = [] then tbl else 
      let stream = get_stream ic oc in
      parse_rif_stream ic oc vntl stream tbl pragma
        
and (parse_rif_stream :  in_channel -> out_channel option -> vntl -> stream -> 
     subst list -> string list -> subst list) =
  fun ic oc vntl stream tbl pragma ->
    if vntl = [] then tbl else
      let tok_list = Stream.npeek 2 stream in
      match tok_list with
	     | [Genlex.Kwd (_,"#"); Genlex.Ident (_,id)] ->
	       if List.mem id pragma then (
		      Stream.junk stream ;
		      Stream.junk stream ;
		      parse_rif_stream ic oc vntl stream tbl pragma
	       ) else (
      (* We skip everything that occurs after a [#], until the next eol. *)
		      Stream.junk stream ;
		      parse_rif_stream ic oc 
              vntl (get_stream ic oc) tbl pragma
	       )
	     | (Genlex.Kwd (_,"#"))::_ ->
          Stream.junk stream ;
          parse_rif_stream ic oc vntl (get_stream ic oc)
            tbl pragma 
	     | (Genlex.Kwd (_,"q"))::_ -> print_string "# bye!\n"; raise Bye
	     | (Genlex.Kwd (_,"#@"))::_ ->
              (* Beginning of multi-line comment. Note that here,
                 unlike the rif format, we ignore multi line pragmas;
                 namely, we handle them as a multi-line comment. *)
          (
            Stream.junk stream ;
            ignore_toks_until_end_of_pragmas ic oc vntl stream tbl pragma 
          )
	     | (Genlex.Float (_,f))::_ ->
          (
            Stream.junk stream ;
           (* Hashtbl.add tbl (Var.name (hd vntl)) (N(F(f))) ; *)
	         let v =
              match snd (hd vntl) with
                | "bool" -> B(f<>0.0)
                | "real" -> F(f)
                | "int"  -> 
                  let i = int_of_float f in
                  print_string "\n*** Warning: type error, ";
		            print_string ((string_of_float f)
				                    ^ " is an real, but an int is expected. I convert it to '"^
                                  (string_of_int i)^"'\n");
                  I(i)
                | e -> 
                  print_string ("\n*** Error: unknown type: "^ e ^ "\n");
                  assert false
	         in
            let tbl = tbl@ [fst (hd vntl), v] in
            parse_rif_stream ic oc (tl vntl) stream tbl pragma 
          )
	     | (Genlex.Int (_,i))::_ -> (
	       Stream.junk stream ;
	       let v =
            match snd (hd vntl) with
              | "bool" -> B(i<>0)
              | "int"  -> I(i)
              | "real" ->
                let f = float_of_int i in
                print_string "\n*** Warning: type error, ";
		          print_string ((string_of_int i)
				                  ^ " is an int, but a real is expected. I convert it to '"^
                                (string_of_float f)^"'\n");
                F(f)
              | e -> 
                print_string ("\n*** Error: unknown type: "^ e ^ "\n");
                assert false
	       in
	       let tbl = tbl @[fst (hd vntl), v] in
          parse_rif_stream ic oc (tl vntl) stream tbl pragma
	     )
	     | (Genlex.Ident (_,b))::_ -> (
	       Stream.junk stream ;
	       let v = if mem b ["f"; "F";"false"] then B(false)
            else if  mem b ["t"; "T";"true"] then B(true)
	         else failwith ("### rif parse error: `" ^ b ^ "' read, where a bool was expected.\n")
	       in
	       let tbl = tbl @ [fst (hd vntl), v] in
          parse_rif_stream ic oc (tl vntl) stream tbl pragma
	     )
	     | [] ->
              (* Eol is is reached; proceed with the next one *)
          parse_rif_stream ic oc vntl (get_stream ic oc)
            tbl pragma 
	     | _ -> failwith ("### rif parse error: not in RIF format.\n")
          
and (ignore_toks_until_end_of_pragmas :   
       in_channel -> out_channel option-> vntl -> stream -> subst list -> string list -> subst list) =
  fun ic oc vntl stream tbl pragma ->
    (* ignore all tokens until "@#" is reached *)
    let tok_opt = Stream.peek stream in
    match tok_opt  with
      | Some(Genlex.Kwd (_,"@#")) ->
        (
          Stream.junk stream ;
          parse_rif_stream ic oc vntl stream tbl pragma
        )
      | Some(_) ->
        (
          Stream.junk stream ;
          ignore_toks_until_end_of_pragmas ic oc vntl stream tbl  pragma 
        )
      | None ->
            (* Eol is is reached; proceed with the next one *)
        (ignore_toks_until_end_of_pragmas ic oc vntl
           (get_stream ic oc) tbl pragma)
        



(*------------------------------------------------------------------------*)
(* exported *)
let (write : out_channel -> string -> unit) =
  fun oc str -> 
    output_string oc str

let (flush : out_channel -> unit) =
  fun oc -> 
    flush oc

(*------------------------------------------------------------------------*)
(* exported *)
let (write_interface : out_channel -> vntl -> vntl -> vntl option -> vntl option -> unit) =
  fun oc in_vars out_vars loc_vars_opt oracle_vars_opt -> 
    let str =  
      (List.fold_left
         (fun acc (vn,vt) ->
            acc ^ "\"" ^ vn ^ "\":" ^ vt ^ " ")
         "#inputs "
         in_vars) ^

        "\n#outputs " ^
        
        (List.fold_left 
           (fun acc (vn,vt) ->
              acc ^ "\"" ^ vn ^ "\":" ^ vt ^ " ")
           ""
           out_vars) ^
        
        (match loc_vars_opt with
           | None -> "\n"
           | Some loc_vars ->  
               ((List.fold_left
                   (fun acc (vn,vt) -> 
                      acc^"\"" ^ vn ^ "\":" ^ vt ^ " ")
                   "\n#locals "
                   loc_vars
                ) ^ "\n")
        )
      ^
        (match oracle_vars_opt with
           | None -> ""
           | Some vars ->  
               ((List.fold_left
                   (fun acc (vn,vt) -> 
                      acc^"\"" ^ vn ^ "\":" ^ vt ^ " ")
                   "#oracle_outputs "
                   vars
                ) ^ "\n")
        )
    in
      write oc str

(*------------------------------------------------------------------------*)
(* exported *)
let (write_outputs : out_channel -> (float -> string) -> vntl -> subst list -> unit) =
  fun oc s2f vntl sl ->
    let str = 
      List.fold_left
        (fun acc (vn,vt) ->
           acc^ (try Data.val_to_string s2f (List.assoc vn sl) 
                 with Not_found -> 
                   print_string ("*** " ^ vn ^ " not found in ");
                   print_string (String.concat "," (List.map (fun (n,_) -> n) sl));
                   flush stdout;
                   assert false) ^ " "
        )
        ""
        vntl
    in
      output_string oc str
